@@ -9,40 +9,28 @@ from __future__ import annotations
 import random
 import abc
 import re
-from typing import List, Dict
+from typing import List, Tuple, Optional
 from subprocess import CalledProcessError, run
-from collections import OrderedDict
+from copy import deepcopy
 
 from .isa_loader import InstructionSet
 from .interfaces import Generator, TestCase, Operand, RegisterOperand, FlagsOperand, \
     MemoryOperand, ImmediateOperand, AgenOperand, LabelOperand, OT, Instruction, BasicBlock, \
-    Function, OperandSpec, InstructionSpec, CondOperand, TargetDesc
-from .util import NotSupportedException, LOGGER
+    Function, OperandSpec, InstructionSpec, CondOperand, Actor, ActorMode, ActorPL, \
+    NotSupportedException
+from .util import Logger
 from .config import CONF
 
 
-# Helpers
-class GeneratorException(Exception):
-    pass
-
-
-class AsmParserException(Exception):
-
-    def __init__(self, line_number, explanation):
-        msg = "Could not parse line " + str(line_number + 1) + "\n  Reason: " + explanation
-        super().__init__(msg)
-
-
-def parser_assert(condition: bool, line_number: int, explanation: str):
-    if not condition:
-        raise AsmParserException(line_number, explanation)
-
-
 # ==================================================================================================
 # Generator Interface
 # ==================================================================================================
 class Pass(abc.ABC):
 
+    def __init__(self) -> None:
+        self.LOG = Logger()
+        super().__init__()
+
     @abc.abstractmethod
     def run_on_test_case(self, test_case: TestCase) -> None:
         pass
@@ -63,16 +51,19 @@ class ConfigurableGenerator(Generator, abc.ABC):
     """
     instruction_set: InstructionSet
     test_case: TestCase
+    parsed_template: Optional[TestCase] = None
     passes: List[Pass]  # set by subclasses
     printer: Printer  # set by subclasses
-    target_desc: TargetDesc  # set by subclasses
+
+    LOG: Logger  # name capitalized to make logging easily distinguishable from the main logic
 
     def __init__(self, instruction_set: InstructionSet, seed: int):
         super().__init__(instruction_set, seed)
-        LOGGER.dbg_gen_instructions(instruction_set.instructions)
+        self.LOG = Logger()
+        self.LOG.dbg_gen_instructions(instruction_set.instructions)
         self.control_flow_instructions = \
             [i for i in self.instruction_set.instructions if i.control_flow]
-        assert self.control_flow_instructions or CONF.max_bb_per_function <= 1, \
+        assert self.control_flow_instructions or CONF.max_successors_per_bb <= 1, \
                "The instruction set is insufficient to generate a test case"
 
         self.non_control_flow_instructions = \
@@ -96,19 +87,30 @@ class ConfigurableGenerator(Generator, abc.ABC):
     def set_seed(self, seed: int) -> None:
         self._state = seed
 
-    def create_test_case(self, asm_file: str, disable_assembler: bool = False) -> TestCase:
-        self.test_case = TestCase(self._state)
-
-        # set seeds
+    def update_seed(self) -> None:
         if self._state == 0:
             self._state = random.randint(1, 1000000)
-            LOGGER.inform("prog_gen",
-                          f"Setting program_generator_seed to random value: {self._state}")
+            self.LOG.inform("prog_gen",
+                            f"Setting program_generator_seed to random value: {self._state}")
         random.seed(self._state)
         self._state += 1
 
+    def create_test_case(self, asm_file: str, disable_assembler: bool = False) -> TestCase:
+        self.test_case = TestCase(self._state)
+        if not asm_file:
+            asm_file = 'generated.asm'
+
+        # set seeds
+        self.update_seed()
+
+        # create actors
+        if len(CONF._actors) != 1:
+            self.LOG.error("Generation of test cases with multiple actors is not yet supported")
+        self.create_actors(self.test_case)
+
         # create the main function
-        func = self.generate_function(".function_main")
+        default_actor = self.test_case.actors["main"]
+        func = self.generate_function(".function_0", default_actor, self.test_case)
 
         # fill the function with instructions
         self.add_terminators_in_function(func)
@@ -116,28 +118,65 @@ class ConfigurableGenerator(Generator, abc.ABC):
 
         # add it to the test case
         self.test_case.functions.append(func)
-        self.test_case.main = func
 
         # process the test case
         for p in self.passes:
             p.run_on_test_case(self.test_case)
 
+        # add symbols to test case
+        self.add_required_symbols(self.test_case)
+
         self.printer.print(self.test_case, asm_file)
         self.test_case.asm_path = asm_file
 
         if disable_assembler:
             return self.test_case
 
-        bin_file = asm_file[:-4] + ".o"
-        self.assemble(asm_file, bin_file)
+        bin_file = asm_file[:-4]
+        obj_file = bin_file + ".o"
+        self.assemble(asm_file, obj_file, bin_file)
         self.test_case.bin_path = bin_file
+        self.test_case.obj_path = obj_file
 
-        self.map_addresses(self.test_case, bin_file)
+        self.get_elf_data(self.test_case, obj_file)
 
         return self.test_case
 
+    def create_test_case_from_template(self, template: str) -> TestCase:
+        self.update_seed()
+
+        if self.parsed_template:
+            test_case = deepcopy(self.parsed_template)
+        else:
+            test_case = self.asm_parser.parse_file(template)
+            self.parsed_template = deepcopy(test_case)
+
+        self.test_case = test_case
+        for func in test_case.functions:
+            for bb in func:
+                for instr in bb:
+                    instr.is_from_template = True
+
+        self.expand_template(test_case)
+        for p in self.passes:
+            p.run_on_test_case(self.test_case)
+
+        asm_file = 'generated.asm'
+        self.printer.print(self.test_case, asm_file)
+        test_case.asm_path = asm_file
+
+        bin_file = asm_file[:-4]
+        obj_file = bin_file + ".o"
+        self.assemble(asm_file, obj_file, bin_file)
+        self.test_case.bin_path = bin_file
+        self.test_case.obj_path = obj_file
+
+        self.test_case.symbol_table = []
+        self.get_elf_data(self.test_case, obj_file)
+        return test_case
+
     @staticmethod
-    def assemble(asm_file: str, bin_file: str) -> None:
+    def assemble(asm_file: str, obj_file: str, bin_file: str) -> None:
         """Assemble the test case into a stripped binary"""
 
         def pretty_error_msg(error_msg):
@@ -156,174 +195,108 @@ class ConfigurableGenerator(Generator, abc.ABC):
             return msg
 
         try:
-            out = run(f"as {asm_file} -o {bin_file}", shell=True, check=True, capture_output=True)
+            out = run(f"as {asm_file} -o {obj_file}", shell=True, check=True, capture_output=True)
         except CalledProcessError as e:
             error_msg = e.stderr.decode()
-            if "Assembler messages:" not in error_msg:
+            if "Assembler messages:" in error_msg:
+                print(pretty_error_msg(error_msg))
+            else:
                 print(error_msg)
-                raise e
-            LOGGER.error(pretty_error_msg(error_msg))
+            raise e
+        finally:
+            pass
+            # run(f"rm {patched_asm_file}", shell=True, check=True)
 
         output = out.stderr.decode()
         if "Assembler messages:" in output:
-            LOGGER.warning("generator", pretty_error_msg(output))
+            print("WARNING: [generator]" + pretty_error_msg(output))
 
+        run(f"cp {obj_file} {bin_file}", shell=True, check=True)
         run(f"strip --remove-section=.note.gnu.property {bin_file}", shell=True, check=True)
         run(f"objcopy {bin_file} -O binary {bin_file}", shell=True, check=True)
 
-    def load(self, asm_file: str) -> TestCase:
-        test_case = TestCase(0)
-        test_case.asm_path = asm_file
+    @abc.abstractmethod
+    def get_elf_data(self, test_case: TestCase, obj_file: str) -> None:
+        pass
 
-        # prepare regexes
-        re_redundant_spaces = re.compile(r"(?<![a-zA-Z0-9]) +")
+    def create_actors(self, test_case: TestCase) -> None:
 
-        # prepare a map of all instruction specs
-        instruction_map: Dict[str, List[InstructionSpec]] = {}
-        for spec in self.instruction_set.instructions:
-            if spec.name in instruction_map:
-                instruction_map[spec.name].append(spec)
-            else:
-                instruction_map[spec.name] = [spec]
+        def pte_properties_to_mask(properties: dict, type_: int) -> int:
+            """
+            Converts a dictionary of PTE properties to a bitmask, later used to set the attributes
+            of faulty pages in the executor.
+            If properties['randomized'] is set to True, each bit has a chance of retaining its
+            default value. Otherwise, the mask is created with the exact values from the dictionary.
+            """
 
-            # add an entry for direct opcodes
-            dummy_spec = InstructionSpec()
-            dummy_spec.name = "OPCODE"
-            dummy_spec.category = "OPCODE"
-            instruction_map["OPCODE"] = [dummy_spec]
+            bits = self.target_desc.pte_bits if type_ == 0 else self.target_desc.epte_bits
 
-        # load the text and clean it up
-        lines = []
-        started = False
-        with open(asm_file, "r") as f:
-            for line in f:
-                # remove extra spaces
-                line = line.strip()
-                line = re_redundant_spaces.sub("", line)
+            # calculate the probability of a bit being set to its default value
+            probability_of_default = 0.0
+            if properties['randomized']:
+                count_non_default = 0
+                for bit_name in bits:
+                    if bits[bit_name][1] != properties[bit_name]:
+                        count_non_default += 1
+                probability_of_default = count_non_default / len(properties)
 
-                # skip comments and empty lines
-                if not line or line[0] in ["", "#", "/"]:
-                    continue
+            # create the mask
+            mask = 0
+            for bit_name in bits:
+                bit_offset, default_value = bits[bit_name]
 
-                # skip footer and header
-                if not started:
-                    started = (line == ".test_case_enter:")
-                    if line[0] != ".":
-                        test_case.num_prologue_instructions += 1
-                    continue
-                if line == ".test_case_exit:":
-                    break
-
-                lines.append(line)
-
-        # set defaults in case the main function is implicit
-        if not lines or not lines[0].startswith(".function_main:"):
-            lines = [".function_main:"] + lines
-
-        # map lines to functions and basic blocks
-        current_function = ""
-        current_bb = ".bb_main.entry"
-        test_case_map: Dict[str, Dict[str, List[str]]] = OrderedDict()
-        for line in lines:
-            # opcode
-            if line[:4] == ".bcd " or line[:5] in [".byte", ".long", ".quad"] \
-               or line[6:] in [".value", ".2byte", ".4byte", ".8byte"]:
-                test_case_map[current_function][current_bb].append("OPCODE")
-                continue
-
-            # instruction
-            if not line.startswith("."):
-                test_case_map[current_function][current_bb].append(line)
-                continue
-
-            # function
-            if line.startswith(".function_"):
-                current_function = line[:-1]
-                test_case_map[current_function] = OrderedDict()
-
-                current_bb = ".bb_" + current_function.removeprefix(".function_") + ".entry"
-                test_case_map[current_function][current_bb] = []
-                continue
-
-            # basic block
-            current_bb = line[:-1]
-            if current_bb not in test_case_map[current_function]:
-                test_case_map[current_function][current_bb] = []
-
-        # parse lines and create their object representations
-        line_id = 1
-        for func_name, bbs in test_case_map.items():
-            # print(func_name)
-            line_id += 1
-            func = Function(func_name)
-            test_case.functions.append(func)
-            if func_name == ".function_main":
-                test_case.main = func
-
-            for bb_name, lines in bbs.items():
-                # print(">>", bb_name)
-                line_id += 1
-                if bb_name.endswith("entry"):
-                    bb = func.entry
-                elif bb_name.endswith("exit"):
-                    bb = func.exit
+                # transform non_executable to executable
+                if bit_name == "non_executable":
+                    p_value = not properties["executable"]
                 else:
-                    bb = BasicBlock(bb_name)
-                    func.insert(bb)
+                    p_value = properties[bit_name]
 
-                terminators_started = False
-                for line in lines:
-                    # print(f"    {line}")
-                    line_id += 1
-                    inst = self.parse_line(line, line_id, instruction_map)
-                    if inst.control_flow and not self.target_desc.is_call(inst):
-                        terminators_started = True
-                        bb.insert_terminator(inst)
-                    else:
-                        parser_assert(not terminators_started, line_id,
-                                      "Terminator not at the end of BB")
-                        bb.insert_after(bb.get_last(), inst)
+                if random.random() < probability_of_default:
+                    p_value = default_value
 
-        # connect basic blocks
-        bb_names = {bb.name.upper(): bb for func in test_case for bb in func}
-        bb_names[".TEST_CASE_EXIT"] = func.exit
-        previous_bb = None
-        for func in test_case:
-            for bb in func:
-                # fallthrough
-                if previous_bb:  # skip the first BB
-                    # there is a fallthrough only if the last terminator is not a direct jump
-                    if not previous_bb.terminators or \
-                       not self.target_desc.is_unconditional_branch(previous_bb.terminators[-1]):
-                        previous_bb.successors.append(bb)
-                previous_bb = bb
+                bit_value = 1 if p_value else 0
+                mask |= bit_value << bit_offset
+            return mask
 
-                # taken branches
-                for terminator in bb.terminators:
-                    for op in terminator.operands:
-                        if isinstance(op, LabelOperand):
-                            successor = bb_names[op.value]
-                            bb.successors.append(successor)
+        for name, desc in CONF._actors.items():
+            # determine the actor mode of execution
+            if desc['mode'] == "host":
+                mode = ActorMode.HOST
+            elif desc['mode'] == "guest":
+                mode = ActorMode.GUEST
+            else:
+                assert False, f"Invalid actor mode: {desc['mode']}"
 
-        bin_file = asm_file[:-4] + ".o"
-        self.assemble(asm_file, bin_file)
-        test_case.bin_path = bin_file
+            if desc['privilege_level'] == "kernel":
+                pl = ActorPL.KERNEL
+            elif desc['privilege_level'] == "user":
+                pl = ActorPL.USER
+            else:
+                assert False, f"Invalid actor privilege_level: {desc['privilege_level']}"
 
-        self.map_addresses(test_case, bin_file)
+            # create the actor
+            if name == "main":
+                actor = test_case.actors["main"]
+            else:
+                id_ = 0  # will be assigned later by the ELF parser
+                actor = Actor(mode, pl, id_, name)
 
-        return test_case
+            # create a PTE mask to be assigned to the faulty area of actors' sandboxes
+            actor.data_properties = pte_properties_to_mask(desc["data_properties"], 0)
+            if actor.mode == ActorMode.GUEST:
+                actor.data_ept_properties = pte_properties_to_mask(desc["data_ept_properties"], 1)
+
+            # assign observer properties (used by non-interference contracts)
+            actor.observer = desc['observer']
+
+            # check for duplicates (this should never be possible, but just in case)
+            assert name not in test_case.actors or test_case.actors[name] == actor, "Duplicate actr"
+
+            # add the actor to the test case
+            test_case.actors[name] = actor
 
     @abc.abstractmethod
-    def parse_line(self, line: str, line_num: int,
-                   instruction_map: Dict[str, List[InstructionSpec]]) -> Instruction:
-        pass
-
-    @abc.abstractmethod
-    def map_addresses(self, test_case: TestCase, bin_file: str) -> None:
-        pass
-
-    @abc.abstractmethod
-    def generate_function(self, name: str) -> Function:
+    def generate_function(self, name: str, owner: Actor, parent: TestCase) -> Function:
         pass
 
     @abc.abstractmethod
@@ -370,6 +343,10 @@ class ConfigurableGenerator(Generator, abc.ABC):
     def generate_cond_operand(self, spec: OperandSpec, _: Instruction) -> Operand:
         pass
 
+    @abc.abstractmethod
+    def expand_template(self, test_case: TestCase) -> None:
+        pass
+
     @abc.abstractmethod
     def add_terminators_in_function(self, func: Function):
         pass
@@ -378,6 +355,10 @@ class ConfigurableGenerator(Generator, abc.ABC):
     def add_instructions_in_function(self, func: Function):
         pass
 
+    @abc.abstractmethod
+    def add_required_symbols(self, test_case: TestCase):
+        pass
+
 
 # ==================================================================================================
 # ISA-independent Generators
@@ -387,7 +368,6 @@ class RandomGenerator(ConfigurableGenerator, abc.ABC):
     Implements an ISA-independent logic of random test case generation.
     Subclasses are responsible for the ISA-specific parts.
     """
-    had_recent_memory_access: bool = False
 
     def __init__(self, instruction_set: InstructionSet, seed: int):
         super().__init__(instruction_set, seed)
@@ -395,21 +375,24 @@ class RandomGenerator(ConfigurableGenerator, abc.ABC):
         self.cond_branches = \
             [i for i in self.control_flow_instructions if i.name.lower() != uncond_name]
 
-    def generate_function(self, label: str):
+    def generate_function(self, label: str, owner: Actor, parent: TestCase):
         """ Generates a random DAG of basic blocks within a function """
-        func = Function(label)
+        func = Function(label, owner)
 
         # Define the maximum allowed number of successors for any BB
         if self.instruction_set.has_conditional_branch:
             max_successors = CONF.max_successors_per_bb if CONF.max_successors_per_bb < 2 else 2
-            min_successors = CONF.min_successors_per_bb \
-                if CONF.min_successors_per_bb < max_successors else max_successors
+            min_successors = CONF.min_successors_per_bb if CONF.min_successors_per_bb < 2 else 2
+            assert min_successors <= max_successors, "min_successors_per_bb > max_successors_per_bb"
         else:
             max_successors = 1
             min_successors = 1
 
         # Create basic blocks
-        node_count = random.randint(CONF.min_bb_per_function, CONF.max_bb_per_function)
+        if CONF.min_bb_per_function == CONF.max_bb_per_function:
+            node_count = CONF.min_bb_per_function
+        else:
+            node_count = random.randint(CONF.min_bb_per_function, CONF.max_bb_per_function)
         func_name = label.removeprefix(".function_")
         nodes = [BasicBlock(f".bb_{func_name}.{i}") for i in range(node_count)]
 
@@ -439,14 +422,14 @@ class RandomGenerator(ConfigurableGenerator, abc.ABC):
                 options.remove(target)
                 current_bb.successors.append(target)
 
-        func.entry.successors = [nodes[0]]
-
-        # Function return
-        if label != ".function_main":
-            func.exit.terminators = [self.get_return_instruction()]
+        # Function returns are not yet supported
+        # hence all functions end with an unconditional jump to the exit
+        func.exit.terminators = [
+            self.get_unconditional_jump_instruction().add_op(LabelOperand(parent.exit.name))
+        ]
 
         # Finalize the function
-        func.insert_multiple(nodes)
+        func.extend(nodes)
         return func
 
     def generate_instruction(self, spec: InstructionSpec) -> Instruction:
@@ -465,26 +448,17 @@ class RandomGenerator(ConfigurableGenerator, abc.ABC):
 
         return inst
 
-    def generate_reg_operand(self, spec: OperandSpec, parent: Instruction) -> Operand:
+    def generate_reg_operand(self, spec: OperandSpec, _: Instruction) -> Operand:
         reg_type = spec.values[0]
-        if reg_type == 'GPR':
+        if reg_type == 'gpr':  # deprecated?
             choices = self.target_desc.registers[spec.width]
-        elif reg_type == "SIMD":
+        elif reg_type == "simd":  # deprecated?
             choices = self.target_desc.simd_registers[spec.width]
         else:
             choices = spec.values
 
-        if not CONF.avoid_data_dependencies:
-            reg = random.choice(choices)
-            return RegisterOperand(reg, spec.width, spec.src, spec.dest)
-
-        if parent.latest_reg_operand and parent.latest_reg_operand.value in choices:
-            return parent.latest_reg_operand
-
         reg = random.choice(choices)
-        op = RegisterOperand(reg, spec.width, spec.src, spec.dest)
-        parent.latest_reg_operand = op
-        return op
+        return RegisterOperand(reg, spec.width, spec.src, spec.dest)
 
     def generate_mem_operand(self, spec: OperandSpec, _: Instruction) -> Operand:
         if spec.values:
@@ -494,21 +468,32 @@ class RandomGenerator(ConfigurableGenerator, abc.ABC):
         return MemoryOperand(address_reg, spec.width, spec.src, spec.dest)
 
     def generate_imm_operand(self, spec: OperandSpec, _: Instruction) -> Operand:
+        # generate bitmask
+        if spec.values and spec.values[0] == "bitmask":
+            # FIXME: this implementation always returns the same bitmask
+            # make it random
+            value = str(pow(2, spec.width) - 2)
+            return ImmediateOperand(value, spec.width)
+
+        # generate from a predefined range
         if spec.values:
-            if spec.values[0] == "bitmask":
-                # FIXME: this implementation always returns the same bitmask
-                # make it random
-                value = str(pow(2, spec.width) - 2)
-            else:
-                assert "[" in spec.values[0], spec.values
-                range_ = spec.values[0][1:-1].split("-")
-                if range_[0] == "":
-                    range_ = range_[1:]
-                    range_[0] = "-" + range_[0]
-                assert len(range_) == 2
-                value = str(random.randint(int(range_[0]), int(range_[1])))
+            assert "[" in spec.values[0], spec.values
+            range_ = spec.values[0][1:-1].split("-")
+            if range_[0] == "":
+                range_ = range_[1:]
+                range_[0] = "-" + range_[0]
+            assert len(range_) == 2
+            value = str(random.randint(int(range_[0]), int(range_[1])))
+            ImmediateOperand(value, spec.width)
+
+        # generate from width
+        if spec.signed:
+            range_min = pow(2, spec.width - 1) * -1
+            range_max = pow(2, spec.width - 1) - 1
         else:
-            value = str(random.randint(pow(2, spec.width - 1) * -1, pow(2, spec.width - 1) - 1))
+            range_min = 0
+            range_max = pow(2, spec.width) - 1
+        value = str(random.randint(range_min, range_max))
         return ImmediateOperand(value, spec.width)
 
     def generate_label_operand(self, spec: OperandSpec, parent: Instruction) -> Operand:
@@ -563,6 +548,40 @@ class RandomGenerator(ConfigurableGenerator, abc.ABC):
         cond = random.choice(list(self.target_desc.branch_conditions))
         return CondOperand(cond)
 
+    def expand_template(self, test_case: TestCase):
+        instr_to_expand: List[Tuple[Instruction, BasicBlock, str]] = []
+        for func in test_case.functions:
+            for bb in func:
+                for instr in bb:
+                    if instr.name == "macro" and instr.operands[0].value == ".random_instructions":
+                        instr_to_expand.append((instr, bb, func.owner.name))
+
+        for inst, bb, a_name in instr_to_expand:
+            operands = inst.operands[1].value.split(".")
+            assert len(operands) >= 3 and len(operands) <= 5
+            n_instr = int(operands[1])
+            n_mem = int(operands[2])
+            predecessor = inst.previous
+
+            # determine the instruction set for this actor
+            blocklist = CONF._actors[a_name]["instruction_blocklist"]
+            non_memory_access_instructions = \
+                [i for i in self.non_memory_access_instructions if i.name not in blocklist]
+            store_instructions = [i for i in self.store_instructions if i.name not in blocklist]
+            load_instruction = [i for i in self.load_instruction if i.name not in blocklist]
+
+            # replace the macro with random instructions
+            bb.delete(inst)
+            for _ in range(n_instr):
+                spec = self._pick_random_instruction_spec(non_memory_access_instructions,
+                                                          store_instructions, load_instruction,
+                                                          n_mem / n_instr)
+                inst = self.generate_instruction(spec)
+                if predecessor:
+                    bb.insert_after(predecessor, inst)
+                else:
+                    bb.insert_before(bb.get_first(), inst)
+
     def add_terminators_in_function(self, func: Function):
 
         def add_fallthrough(bb: BasicBlock, destination: BasicBlock):
@@ -601,38 +620,35 @@ class RandomGenerator(ConfigurableGenerator, abc.ABC):
 
     def add_instructions_in_function(self, func: Function):
         # evenly fill all BBs with random instructions
-        basic_blocks_to_fill = func.get_all()[1:-1]
+        bb_list = func[:]
         for _ in range(0, CONF.program_size):
-            bb = random.choice(basic_blocks_to_fill)
-            spec = self._pick_random_instruction_spec()
+            bb = random.choice(bb_list)
+            spec = self._pick_random_instruction_spec(self.non_memory_access_instructions,
+                                                      self.store_instructions,
+                                                      self.load_instruction,
+                                                      CONF.avg_mem_accesses / CONF.program_size)
             inst = self.generate_instruction(spec)
             bb.insert_after(bb.get_last(), inst)
 
-    def _pick_random_instruction_spec(self) -> InstructionSpec:
+    def _pick_random_instruction_spec(self,
+                                      non_memory_access_instructions: List,
+                                      store_instructions: List,
+                                      load_instructions: List,
+                                      memory_access_probability: float = 0.0) -> InstructionSpec:
         # ensure the requested avg. number of mem. accesses
-        search_for_memory_access = False
-        memory_access_probability = CONF.avg_mem_accesses / CONF.program_size
-        if CONF.generate_memory_accesses_in_pairs:
-            memory_access_probability = 1 if self.had_recent_memory_access else \
-                (CONF.avg_mem_accesses / 2) / (CONF.program_size - CONF.avg_mem_accesses / 2)
+        search_for_memory_access = random.random() < memory_access_probability
+        if not search_for_memory_access:
+            return random.choice(non_memory_access_instructions)
 
-        if random.random() < memory_access_probability:
-            search_for_memory_access = True
-            self.had_recent_memory_access = not self.had_recent_memory_access
-
-        if self.store_instructions:
+        if store_instructions:
             search_for_store = random.random() < 0.5  # 50% probability of stores
         else:
             search_for_store = False
 
-        # select a random instruction spec for generation
-        if not search_for_memory_access:
-            return random.choice(self.non_memory_access_instructions)
-
         if search_for_store:
-            return random.choice(self.store_instructions)
+            return random.choice(store_instructions)
 
-        return random.choice(self.load_instruction)
+        return random.choice(load_instructions)
 
     @abc.abstractmethod
     def get_return_instruction(self) -> Instruction:
@@ -641,3 +657,20 @@ class RandomGenerator(ConfigurableGenerator, abc.ABC):
     @abc.abstractmethod
     def get_unconditional_jump_instruction(self) -> Instruction:
         pass
+
+    def add_required_symbols(self, test_case: TestCase):
+        # add measurement_start and measurement_end symbols
+        func_main = test_case.functions[0]
+        assert func_main.owner == test_case.actors["main"]
+
+        bb_first = func_main[0]
+        instr = Instruction("macro", category="MACRO") \
+            .add_op(LabelOperand(".measurement_start")) \
+            .add_op(LabelOperand(".noarg"))
+        bb_first.insert_before(bb_first.get_first(), instr)
+
+        bb_last = func_main.exit
+        instr = Instruction("macro", category="MACRO") \
+            .add_op(LabelOperand(".measurement_end")) \
+            .add_op(LabelOperand(".noarg"))
+        bb_last.insert_after(bb_last.get_last(), instr)
