@@ -7,17 +7,18 @@ SPDX-License-Identifier: MIT
 from __future__ import annotations
 
 import shutil
+import xxhash
 from typing import List, Dict, Tuple, Optional, NamedTuple
 from collections import defaultdict
 from abc import ABC, abstractmethod
 import numpy as np
 from enum import Enum
 
-from .config import CONF
+PAGE_SIZE = 4096
 
 
 # ==================================================================================================
-# Components of a Test Case
+# Instruction Specifications
 # ==================================================================================================
 class OT(Enum):
     """Operand Type"""
@@ -37,6 +38,7 @@ class OperandSpec:
     values: List[str]
     type: OT
     width: int
+    signed: bool = True
     src: bool
     dest: bool
 
@@ -55,6 +57,243 @@ class OperandSpec:
         return f"{self.values}"
 
 
+class InstructionSpec:
+    name: str
+    operands: List[OperandSpec]
+    implicit_operands: List[OperandSpec]
+    category: str
+    control_flow = False
+
+    has_mem_operand = False
+    has_write = False
+    has_magic_value: bool = False
+
+    def __init__(self):
+        self.operands = []
+        self.implicit_operands = []
+
+    def __str__(self):
+        ops = ""
+        for o in self.operands:
+            ops += str(o) + " "
+        return f"{self.name} {ops}"
+
+
+class MacroSpec(NamedTuple):
+    type_: SymbolType
+    name: str
+    args: Tuple[str, str, str, str]
+
+
+# ==================================================================================================
+# Sandbox layout and constants
+# ==================================================================================================
+# Note: the constants *must* be identical to the actor_data_t and actor_code_t in executor
+# - Data layout:
+#   +-------------------+ (page-aligned)
+#   | OVERFLOW_PAD_SIZE |
+#   +-------------------+
+#   | REG_INIT_AREA_SIZE|
+#   +-------------------+ (page-aligned)
+#   | FAULTY_AREA_SIZE  |
+#   +-------------------+ (page-aligned)
+#   | MAIN_AREA_SIZE    |
+#   +-------------------+ (page-aligned)
+#   | UNDERFLOW_PAD_SIZE|
+#   +-------------------+
+#   | MACRO_STACK_SIZE  |
+#   +-------------------+ (page-aligned)
+MACRO_STACK_SIZE = 64
+UNDERFLOW_PAD_SIZE = (PAGE_SIZE - MACRO_STACK_SIZE)
+MAIN_AREA_SIZE = PAGE_SIZE
+FAULTY_AREA_SIZE = PAGE_SIZE
+GPR_SUBREGION_SIZE = 64  # 8 64-bit GPRs
+SIMD_SUBREGION_SIZE = 256  # 8 256-bit YMMs
+REG_INIT_AREA_SIZE = (GPR_SUBREGION_SIZE + SIMD_SUBREGION_SIZE)
+OVERFLOW_PAD_SIZE = (PAGE_SIZE - REG_INIT_AREA_SIZE)
+SANDBOX_DATA_SIZE = MAIN_AREA_SIZE + FAULTY_AREA_SIZE + 2 * PAGE_SIZE
+
+
+def get_sandbox_addr(sandbox_base: int, name: str):
+    name_to_addr = {
+        "start": lambda x: x - UNDERFLOW_PAD_SIZE - MACRO_STACK_SIZE,
+        "macro_stack": lambda x: x - UNDERFLOW_PAD_SIZE - MACRO_STACK_SIZE,
+        "underflow_pad": lambda x: x - UNDERFLOW_PAD_SIZE,
+        "main": lambda x: x,
+        "faulty": lambda x: x + MAIN_AREA_SIZE,
+        "sp": lambda x: x + MAIN_AREA_SIZE - 8,
+        "gpr": lambda x: x + MAIN_AREA_SIZE + FAULTY_AREA_SIZE,
+        "simd": lambda x: x + MAIN_AREA_SIZE + FAULTY_AREA_SIZE + GPR_SUBREGION_SIZE,
+        "reg_init": lambda x: x + MAIN_AREA_SIZE + FAULTY_AREA_SIZE,
+        "overflow_pad": lambda x: x + MAIN_AREA_SIZE + FAULTY_AREA_SIZE + REG_INIT_AREA_SIZE,
+    }
+    if name not in name_to_addr:
+        raise Exception(f"Invalid sandbox region name: {name}")
+    return name_to_addr[name](sandbox_base)
+
+
+# - Code layout:
+SANDBOX_CODE_SIZE = 3 * PAGE_SIZE
+
+# ==================================================================================================
+# Actors
+# ==================================================================================================
+ActorID = int
+ActorName = str
+
+
+class ActorMode(Enum):
+    HOST = 0
+    GUEST = 1
+
+
+class ActorPL(Enum):
+    KERNEL = 0
+    USER = 1
+
+
+class ElfSection(NamedTuple):
+    id_: int  # section id; will match the actor id if the actor IDs are ordered and contiguous
+    offset: int
+    size: int
+
+
+class Actor:
+    name: ActorName
+    mode: ActorMode
+    privilege_level: ActorPL
+    id_: ActorID
+    elf_section: Optional[ElfSection] = None
+    data_properties: int = 0
+    data_ept_properties: int = 0
+    code_properties: int = 0  # unused so far
+    observer: bool = False
+
+    def __init__(self, mode: ActorMode, pl: ActorPL, id_: ActorID, name: ActorName) -> None:
+        self.mode = mode
+        self.privilege_level = pl
+        self.id_ = id_
+        self.name = name
+
+
+# ==================================================================================================
+# Input
+# ==================================================================================================
+# InputFragment data type represents the input for a single actor. This array is designed to be
+# easily serialized and deserialized into/from a binary file.
+# InputFragment is a fixed-size array of 64-bit unsigned integers,
+# structured into 2 sub-arrays representing memory (`main` and `faulty`),
+# and 2 sub-arrays representing CPU registers (`gpr` and `simd`).
+# The array is used to initialize sandbox memory and CPU registers.
+#
+# Ordering of GPRs:  RAX, RBX, RCX, RDX, RSI, RDI, FLAGS, (last 8 bytes unused)
+# Ordering of SIMD registers: YMM0, YMM1, ..., YMM7
+InputFragment = np.dtype(
+    [
+        ('main', np.uint64, MAIN_AREA_SIZE // 8),
+        ('faulty', np.uint64, FAULTY_AREA_SIZE // 8),
+        ('gpr', np.uint64, GPR_SUBREGION_SIZE // 8),
+        ('simd', np.uint64, SIMD_SUBREGION_SIZE // 8),
+        ('padding', np.uint64, (4096 - REG_INIT_AREA_SIZE) // 8),
+    ],
+    align=False,
+)
+
+
+class Input(np.ndarray):
+    """
+    This class represents a single input to a test case. It is a fixed-size array of
+    64-bit unsigned integers, with a few addition methods for convenience. The array
+    is used to initialize the sandbox memory and the CPU registers.
+
+    The number of elements in Input is equal to the number of actors multiplied by
+    the number of elements in InputFragment, i.e.,
+        Input.size = n_actors * InputFragment.size
+    """
+    seed: int = 0
+    data_size: int
+
+    def __init__(self, n_actors: int = 1) -> None:
+        pass  # unreachable; defined only for type checking
+
+    def __new__(cls, n_actors: int = 1):
+        obj = super().__new__(cls, (n_actors,), InputFragment, None, 0, None, None)  # type: ignore
+        obj.data_size = (MAIN_AREA_SIZE + FAULTY_AREA_SIZE + REG_INIT_AREA_SIZE) // 8
+        return obj
+
+    def __array_finalize__(self, obj):
+        if obj is None:
+            return
+
+    def __hash__(self) -> int:  # type: ignore
+        # hash of input is hash of input data, registers and memory
+        h = hash(self.tobytes())
+        return h
+
+    def get_simd128_registers(self, actor_id: int):
+        vals = []
+        for i in range(0, InputFragment['simd'].shape[actor_id], 2):
+            vals.append(int(self[0]['simd'][i + 1]) << 64 | int(self[0]['simd'][i]))
+        return vals
+
+    def __str__(self):
+        return str(self.seed)
+
+    def __repr__(self):
+        return str(self.seed)
+
+    def linear_view(self, actor_id) -> np.ndarray:
+        return self[actor_id].view((np.uint64, self[actor_id].itemsize // 8))
+
+    def save(self, path: str) -> None:
+        with open(path, 'wb') as f:
+            f.write(self.tobytes())
+
+    def load(self, path: str) -> None:
+        with open(path, 'rb') as f:
+            contents = np.fromfile(f, dtype=np.uint64)
+            n_actors = self.shape[0]
+            for actor_id in range(n_actors):
+                actor_start = actor_id * self.itemsize // 8
+                actor_end = actor_start + self.itemsize // 8
+                self.linear_view(actor_id)[:] = contents[actor_start:actor_end]
+
+
+class InputTaint(Input):
+    """
+    This class represents a taint vector for an input. It is a fixed-size array of
+    boolean values, with a few addition methods for convenience. The array is used
+    to indicate which input elements influence contract traces. The number of
+    elements in InputTaint is identical to Input class.
+
+    Each element is an boolean value: When it is True, the corresponding element of
+    the input impacts the contract trace.
+    """
+
+    def __new__(cls, n_actors: int = 1):
+        obj = super().__new__(cls, n_actors)  # type: ignore
+        obj.fill(False)
+        return obj
+
+    def __init__(self, n_actors: int = 1) -> None:
+        pass  # unreachable; defined only for type checking
+
+
+# ==================================================================================================
+# Test Cases
+# ==================================================================================================
+SymbolType = int
+SymbolOffset = int
+MacroArgument = int
+
+
+class Symbol(NamedTuple):
+    aid: ActorID
+    offset: SymbolOffset
+    type_: SymbolType
+    arg: MacroArgument
+
+
 class Operand(ABC):
     value: str
     type: OT
@@ -81,14 +320,14 @@ class RegisterOperand(Operand):
 
     def __init__(self, value: str, width: int, src: bool, dest: bool):
         self.width = width
-        super().__init__(value.upper(), OT.REG, src, dest)
+        super().__init__(value.lower(), OT.REG, src, dest)
 
 
 class MemoryOperand(Operand):
 
     def __init__(self, address: str, width: int, src: bool, dest: bool):
         self.width = width
-        super().__init__(address.upper(), OT.MEM, src, dest)
+        super().__init__(address.lower(), OT.MEM, src, dest)
 
 
 class ImmediateOperand(Operand):
@@ -108,7 +347,7 @@ class AgenOperand(Operand):
 
     def __init__(self, value: str, width: int):
         self.width = width
-        super().__init__(value.upper(), OT.AGEN, True, False)
+        super().__init__(value.lower(), OT.AGEN, True, False)
 
 
 class FlagsOperand(Operand):
@@ -145,6 +384,9 @@ class FlagsOperand(Operand):
     def get_write_flags(self) -> List[str]:
         return self._get_flag_list(['w', 'r/w', 'r/cw'])
 
+    def get_overwrite_flags(self) -> List[str]:
+        return self._get_flag_list(['w'])
+
     def get_undef_flags(self) -> List[str]:
         return self._get_flag_list(['undef'])
 
@@ -162,41 +404,45 @@ class CondOperand(Operand):
         super().__init__(value, OT.COND, True, False)
 
 
-class InstructionSpec:
-    name: str
-    operands: List[OperandSpec]
-    implicit_operands: List[OperandSpec]
-    category: str
-    control_flow = False
-
-    has_mem_operand = False
-    has_write = False
-    has_magic_value: bool = False
-
-    def __init__(self):
-        self.operands = []
-        self.implicit_operands = []
-
-    def __str__(self):
-        ops = ""
-        for o in self.operands:
-            ops += str(o) + " "
-        return f"{self.name} {ops}"
-
-
 class Instruction:
     name: str
+    """ name: The name of the instruction without any operands """
     operands: List[Operand]
+    """ operands: List of explicit operands of the instruction """
     implicit_operands: List[Operand]
+    """ implicit_operands: List of implicit operands, which are not explicitly specified in the
+    instruction but are used by the instruction. For example, flags operand in x86 instructions """
     category: str
-    control_flow = False
+    """ category: The category of the instruction, e.g., BASE-BINARY. The keyword matches
+    the category in the instruction set description file (typically called base.json)"""
+    control_flow: bool = False
+    """ control_flow: If True, the instruction is a control flow instruction
+    (branch, call, return, etc.) """
 
     next: Optional[Instruction] = None
+    """ next: Next instruction in the double-linked list of instructions """
     previous: Optional[Instruction] = None
-    is_instrumentation: bool
+    """ previous: Previous instruction in the double-linked list of instructions """
 
-    # TODO: remove latest_reg_operand from this class. It belongs in the generator
-    latest_reg_operand: Optional[Operand] = None  # for avoiding dependencies
+    is_instrumentation: bool
+    """ is_instrumentation: If True, the instruction is an instrumentation instruction,
+    which means that it was inserted by the generator to prevent faults or false positives """
+    is_from_template: bool = False
+    """ is_from_template: If True, the instruction was directly copied from the template rather
+    then being automatically created by the generator """
+    is_noremove: bool = False
+    """ is_noremove: If True, the instruction should be skipped while doing minimization passes """
+    section_id: int = 0
+    """ section_id: The ID of the section in the object file where the instruction is located """
+    section_offset: int = 0
+    """ section_offset: The section offset of the instruction in the object file """
+    line_num: int = 0
+    """ line_num: The line number in the source (assembly) file where the instruction is located """
+    size: int = 0
+    """ size: The size of the instruction in bytes, after it has been assembled """
+    _inst_brief: str = ""
+    """ _inst_brief: A brief representation of the instruction,
+    used for hashing and for debug messages """
 
     def __init__(self, name: str, is_instrumentation=False, category="", control_flow=False):
         self.name = name
@@ -347,6 +593,29 @@ class Instruction:
 
         return res
 
+    def get_brief(self) -> str:
+        if self._inst_brief:
+            return self._inst_brief
+
+        brief = self.name
+        for o in self.operands:
+            if o.type == OT.REG:
+                brief += f" R{o.width}"
+            elif o.type == OT.MEM:
+                brief += f" M{o.width}"
+            elif o.type == OT.IMM:
+                brief += f" I{o.width}"
+            elif o.type == OT.LABEL:
+                brief += " L"
+            elif o.type == OT.AGEN:
+                brief += f" A{o.width}"
+            elif o.type == OT.FLAGS:
+                brief += " F"
+            elif o.type == OT.COND:
+                brief += " C"
+        self._inst_brief = brief
+        return brief
+
 
 class BasicBlock:
     name: str
@@ -448,18 +717,16 @@ class BasicBlock:
 
 class Function:
     name: str
+    owner: Actor
     _all_bb: List[BasicBlock]
-    entry: BasicBlock
     exit: BasicBlock
+    obj_file_offset: int = 0
 
-    def __init__(self, name):
+    def __init__(self, name: str, owner: Actor):
         self.name = name
-
-        # create entry and exit points for the function
-        stripped_name = name.lstrip(".function_")
-        self.entry = BasicBlock(f".bb_{stripped_name}.entry")
-        self.exit = BasicBlock(f".bb_{stripped_name}.exit")
-        self._all_bb = [self.entry, self.exit]
+        self.owner = owner
+        self.exit = BasicBlock(f".exit_{name.removeprefix('.function_')}")
+        self._all_bb = []
 
     def __len__(self):
         return len(self._all_bb)
@@ -468,145 +735,134 @@ class Function:
         for bb in self._all_bb:
             yield bb
 
-    def insert(self, bb: BasicBlock):
-        self._all_bb = self._all_bb[0:-1]
+    def __getitem__(self, item):
+        return self._all_bb[item]
+
+    def append(self, bb: BasicBlock):
         self._all_bb.append(bb)
-        self._all_bb.append(self.exit)
 
-    def insert_multiple(self, bb_list: List[BasicBlock]):
-        self._all_bb = self._all_bb[0:-1]
-        self._all_bb += bb_list
-        self._all_bb.append(self.exit)
+    def extend(self, bb_list: List[BasicBlock]):
+        self._all_bb.extend(bb_list)
 
-    def get_all(self):
-        return self._all_bb
+    def get_first_bb(self):
+        return self._all_bb[0]
 
 
 class TestCase:
     asm_path: str = ''
+    obj_path: str = ''
     bin_path: str = ''
-    main: Function
+    actors: Dict[ActorName, Actor]
     functions: List[Function]
-    address_map: Dict[int, Instruction]
-    num_prologue_instructions: int = 0
-    faulty_pte: PageTableModifier
+    address_map: Dict[ActorID, Dict[int, Instruction]]
     seed: int
+    exit: BasicBlock
+    symbol_table: List[Symbol]
 
     def __init__(self, seed: int):
         self.seed = seed
+        self.actors = {"main": Actor(ActorMode.HOST, ActorPL.KERNEL, 0, "main")}
         self.functions = []
         self.address_map = {}
-        self.faulty_pte = PageTableModifier()
+        self.symbol_table = []
+        self.exit = BasicBlock(".test_case_exit")
 
     def __iter__(self):
         for func in self.functions:
             yield func
 
+    def get_function_by_name(self, name: str) -> Function:
+        for func in self.functions:
+            if func.name == name:
+                return func
+        raise Exception(f"ERROR: Function {name} not found")
+
+    def get_actor_by_id(self, aid: ActorID) -> Actor:
+        for actor in self.actors.values():
+            if actor.id_ == aid:
+                return actor
+        raise Exception(f"ERROR: Actor {aid} not found")
+
     def save(self, path: str) -> None:
         shutil.copy2(self.asm_path, path)
 
 
-class PageTableModifier(NamedTuple):
-    mask_set: int = 0x0
-    mask_clear: int = 0xffffffffffffffff
-
-
 # ==================================================================================================
-# Custom Data Types
+# Traces
 # ==================================================================================================
-CTrace = int
-HTrace = int
 InputID = int
-CombinedHTrace = int
 
 
-class Input(np.ndarray):
+class CTrace:
     """
-    A class representing a single input to a test case.
-    It is a fixed-size array of 64-bit unsigned integers, with a few addition
-    methods for convenience.
-    The array is used to initialize the sandbox memory and the CPU registers.
-    The array layout is:
-
-    +----------------------+
-    |   Register Values    | Conf.input_register_region_size
-    +----------------------+
-    |                      |
-    |                      | Conf.input_faulty_region_size
-    | Assist Region Values |
-    +----------------------+
-    |                      |
-    |                      | Conf.input_main_region_size
-    |  Main Region Values  |
-    +----------------------+
-
-    The ordering of registers:  RAX, RBX, RCX, RDX, RSI, RDI, FLAGS
+    CTrace is a class that represents a contract trace. It is a container for a list of integers
+    that represent raw trace collected from the model and a hash of the trace. The hash is used to
+    compare traces for equality.
     """
-    seed: int = 0
-    data_size: int = 0
-    register_start: int = 0
+    raw: List[int]
+    hash_: int
 
-    def __init__(self) -> None:
-        pass  # unreachable; defined only for type checking
+    def __init__(self, raw_trace: List) -> None:
+        self.raw = raw_trace
+        self.hash_ = xxhash.xxh64(str(raw_trace), seed=0).intdigest()
 
-    def __new__(cls):
-        data_size = (CONF.input_main_region_size + CONF.input_faulty_region_size
-                     + CONF.input_register_region_size) // 8
-        aligned_size = data_size + (4096 - CONF.input_register_region_size) // 8
-        obj = super().__new__(cls, (aligned_size,), np.uint64, None, 0, None, None)  # type: ignore
-        obj.data_size = data_size
-        obj.register_start = data_size - CONF.input_register_region_size // 8
+    def __eq__(self, other):
+        return self.hash_ == other.hash_
 
-        # fill the input with zeroes initially before returning, to ensure the
-        # 'padding' bytes (created by using 'aligned_size' rather than
-        # 'data_size') deterministic across separate runs
-        obj.fill(0)
-        return obj
+    def __lt__(self, other):
+        return self.hash_ < other.hash_
 
-    def __array_finalize__(self, obj):
-        if obj is None:
-            return
+    def __gt__(self, other):
+        return self.hash_ > other.hash_
 
-    def get_registers(self):
-        return list(self[self.register_start:self.data_size - 1])
-
-    def get_memory(self):
-        return self[0:self.register_start]
+    def __len__(self):
+        return len(self.raw)
 
     def __str__(self):
-        return str(self.seed)
+        return str(self.hash_)
 
-    def __repr__(self):
-        return str(self.seed)
+    def __hash__(self) -> int:
+        return self.hash_
 
-    def save(self, path: str) -> None:
-        with open(path, 'wb') as f:
-            f.write(self.tobytes())
-
-    def load(self, path: str) -> None:
-        with open(path, 'rb') as f:
-            contents = np.fromfile(f, dtype=np.uint64)
-            self[:] = contents
+    @classmethod
+    def get_null(cls):
+        """ Get a dummy CTrace object with empty contract trace """
+        return cls([])
 
 
-class InputTaint(np.ndarray):
+class HTrace:
     """
-    An array that represents which input elements influence contract traces.
-    The number of elements in InputTaint is identical to Input class.
-    Each element is an boolean value: When it is True, the corresponding element
-    of the input impacts the contract trace.
+    HTrace is a class that represents a hardware trace. It is a container for a list of integers
+    that represents hardware traces collected from the executor over multiple repeated runs
+    with the same program and input. It also contains a hash of the trace and a list of values for
+    performance counters collected together with the hardware trace. The hash is used to compare
+    traces for equality.
     """
-    register_start: int = 0
+    raw: List[int]
+    hash_: int
+    perf_counters: np.ndarray
+    perf_counters_max: List[int]
 
-    def __init__(self) -> None:
-        pass  # unreachable; defined only for type checking
+    def __init__(self, trace_list: List[int], perf_counters: Optional[np.ndarray] = None) -> None:
+        self.raw = trace_list
+        self.hash_ = xxhash.xxh64(str(trace_list), seed=0).intdigest()
+        if perf_counters is None:
+            self.perf_counters = np.array([0, 0, 0, 0, 0])
+            self.perf_counters_max = [0, 0, 0, 0, 0]
+        else:
+            self.perf_counters = perf_counters
+            self.perf_counters_max = max(perf_counters, key=lambda x: x[0])
 
-    def __new__(cls):
-        size = (CONF.input_main_region_size + CONF.input_faulty_region_size
-                + CONF.input_register_region_size) // 8
-        obj = super().__new__(cls, (size,), bool, None, 0, None, None)  # type: ignore
-        obj.register_start = (CONF.input_main_region_size + CONF.input_faulty_region_size) // 8
-        return obj
+    def __eq__(self, other):
+        return self.hash_ == other.hash_
+
+    def __len__(self):
+        return len(self.raw)
+
+    @classmethod
+    def get_null(cls):
+        """ Get a dummy HTrace object with empty hardware trace and zeros for perf counters """
+        return cls([])
 
 
 class Measurement(NamedTuple):
@@ -616,39 +872,48 @@ class Measurement(NamedTuple):
     htrace: HTrace
 
 
-HTraceGroup = List[Measurement]
-HTraceMap = Dict[HTrace, HTraceGroup]
-
-
 class EquivalenceClass:
     ctrace: CTrace
     measurements: List[Measurement]
-    htrace_map: HTraceMap
+    htrace_groups: List[List[Measurement]]
+    """ htrace_groups: a list of htrace groups; each group is a list of measurements that produced
+    the same htrace (or a equivalent htraces under the current analyser). """
+
     MOD2P64 = pow(2, 64)
 
-    def __init__(self) -> None:
-        self.measurements = []
-
-    def __str__(self):
-        s = f"Size: {len(self.measurements)}\n"
-        s += f"Ctrace:\n" \
-             f"{self.ctrace % self.MOD2P64:064b} [ns]\n" \
-             f"{(self.ctrace >> 64) % self.MOD2P64:064b} [s]\n"
-        s += "Htraces:\n"
-        for h in self.htrace_map.keys():
-            s += f"{h:064b}\n"
-        s = s.replace("0", "_").replace("1", "^")
-        return s
+    def __init__(self, ctrace: CTrace, measurements: List[Measurement],
+                 htrace_groups: List[List[Measurement]]) -> None:
+        self.ctrace = ctrace
+        self.measurements = measurements
+        self.htrace_groups = htrace_groups
 
     def __len__(self):
         return len(self.measurements)
 
-    def build_htrace_map(self) -> None:
-        """ group inputs by htraces """
-        groups = defaultdict(list)
-        for measurement in self.measurements:
-            groups[measurement.htrace].append(measurement)
-        self.htrace_map = groups
+
+class Violation(EquivalenceClass):
+    """ Violation is a special type of equivalence class that represents a violation of a contract.
+    It is a container for a list of measurements that triggered the violation, a list of groups of
+    htraces, and a sequence of inputs that triggered the violation. """
+
+    ctrace: CTrace
+    measurements: List[Measurement]
+    htrace_groups: List[List[Measurement]]
+    """ htrace_groups: a list of htrace groups; each group is a list of measurements that produced
+    the same htrace (or a equivalent htraces under the current analyser). """
+    input_sequence: List[Input]
+    """ input_sequence: the complete sequence of inputs that triggered the violation """
+
+    def __init__(self, eq_cls: EquivalenceClass, inputs: List[Input]) -> None:
+        self.measurements = eq_cls.measurements
+        self.ctrace = eq_cls.ctrace
+        self.htrace_groups = eq_cls.htrace_groups
+        self.input_sequence = inputs
+
+    @classmethod
+    def from_measurements(cls, ctrace: CTrace, measurements: List[Measurement],
+                          htrace_groups: List[List[Measurement]], inputs: List[Input]):
+        return cls(EquivalenceClass(ctrace, measurements, htrace_groups), inputs)
 
 
 # Execution Tracing
@@ -670,7 +935,8 @@ ExecutionTrace = List[TracedInstruction]
 # Interfaces of Modules
 # ==================================================================================================
 class InstructionSetAbstract(ABC):
-    instructions: List[InstructionSpec] = []
+    instructions: List[InstructionSpec]
+    instruction_unfiltered: List[InstructionSpec]
     has_unconditional_branch: bool = False
     has_conditional_branch: bool = False
     has_indirect_branch: bool = False
@@ -682,12 +948,24 @@ class InstructionSetAbstract(ABC):
         pass
 
 
+class CPUDesc(NamedTuple):
+    vendor: str
+    model: str
+    family: str
+    stepping: str
+
+
 class TargetDesc(ABC):
     register_sizes: Dict[str, int]
     registers: Dict[int, List[str]]
     simd_registers: Dict[int, List[str]]
     branch_conditions: Dict[str, List[str]]
-    gpr_normalized: Dict[str, str]
+    reg_normalized: Dict[str, str]
+    reg_denormalized: Dict[str, Dict[int, str]]
+    macro_specs: Dict[str, MacroSpec]
+    pte_bits: Dict[str, Tuple[int, bool]]
+    epte_bits: Dict[str, Tuple[int, bool]]
+    cpu_desc: CPUDesc
 
     @staticmethod
     @abstractmethod
@@ -700,8 +978,23 @@ class TargetDesc(ABC):
         pass
 
 
+class AsmParser(ABC):
+    """ Class responsible for parsing (and optionally patching) assembly files
+    and producing TestCases from them """
+
+    @abstractmethod
+    def parse_file(self, asm_file: str) -> TestCase:
+        """
+        Read a test case from a file and create a complete TestCase object based on it.
+        Used instead of create_test_case when Revizor works with a user-provided test case.
+        """
+        pass
+
+
 class Generator(ABC):
     instruction_set: InstructionSetAbstract
+    target_desc: TargetDesc
+    asm_parser: AsmParser
     _state: int = 0
 
     def __init__(self, instruction_set: InstructionSetAbstract, seed: int):
@@ -727,31 +1020,45 @@ class Generator(ABC):
     @abstractmethod
     def create_test_case(self, path: str, disable_assembler: bool = False) -> TestCase:
         """
-        Create a simple test case with a single BB
+        Generate a random test case base on the config options.
         Run instrumentation passes and print the result into a file
         """
         pass
 
     @abstractmethod
-    def load(self, asm_file: str) -> TestCase:
+    def create_test_case_from_template(self, template: str) -> TestCase:
         """
-        Read a test case from a file and create a complete TestCase object based on it.
-        Used instead of create_test_case when Revizor works with a user-provided test case.
+        Generate a test case based on a template by expanding RANDOM_* macros.
+        Run instrumentation passes and print the result into a file
         """
         pass
 
     @staticmethod
     @abstractmethod
-    def assemble(asm_file: str, bin_file: str) -> None:
+    def assemble(asm_file: str, obj_file: str, bin_file: str) -> None:
+        """
+        Assemble an assembly file into an object file and creates a stripped binary
+        """
         pass
 
     @abstractmethod
-    def create_pte(self, test_case: TestCase) -> None:
+    def get_elf_data(self, test_case: TestCase, obj_file: str) -> None:
+        """
+        Extract ELF data from an object file and populate the test case
+        """
+        pass
+
+    @abstractmethod
+    def create_actors(self, test_case: TestCase) -> None:
+        """
+        Create actors for the test case based on the description in CONF._actors
+        """
         pass
 
 
 class InputGenerator(ABC):
     _state: int = 0
+    n_actors: int = 1
 
     def __init__(self, seed: int):
         self.set_seed(seed)
@@ -770,6 +1077,10 @@ class InputGenerator(ABC):
         """
         return self._state
 
+    @abstractmethod
+    def reset_boosting_state(self) -> None:
+        pass
+
     @abstractmethod
     def generate(self, count: int) -> List[Input]:
         pass
@@ -787,62 +1098,39 @@ class InputGenerator(ABC):
         pass
 
 
-class Coverage(ABC):
-    instruction_set: InstructionSetAbstract
-
-    def __init__(self, instruction_set: InstructionSetAbstract, executor: Executor, model: Model,
-                 analyser: Analyser):
-        self.instruction_set = instruction_set
-        executor.set_coverage(self)
-        model.set_coverage(self)
-        analyser.set_coverage(self)
-        super().__init__()
-
-    @abstractmethod
-    def get(self) -> int:
-        pass
-
-    def get_brief(self) -> str:
-        return ""
-
-    @abstractmethod
-    def load_test_case(self, test_case: TestCase):
-        pass
-
-    @abstractmethod
-    def model_hook(self, feedback):
-        pass
-
-    @abstractmethod
-    def executor_hook(self, feedback):
-        pass
-
-    @abstractmethod
-    def analyser_hook(self, feedback):
-        pass
-
-
 class Tracer(ABC):
     trace: List
 
     @abstractmethod
-    def get_contract_trace(self, model: Model) -> CTrace:
+    def produce_trace(self, model: Model) -> CTrace:
         pass
 
-    def get_contract_trace_full(self) -> List[int]:
-        return self.trace
-
 
 class Model(ABC):
-    coverage: Optional[Coverage] = None
     sandbox_base: int = 0
     code_start: int = 0
-    lower_overflow_base: int = 0
-    upper_overflow_base: int = 0
+    code_end: int = 0
+    data_start: int = 0
+    data_end: int = 0
     tracer: Tracer
+    instruction_coverage: Dict[str, int]
+    is_speculative_contract: bool = False
 
-    @abstractmethod
-    def __init__(self, sandbox_base: int, code_base: int):
+    mismatch_check_mode: bool = False
+    """ mismatch_check_mode: If True, the model will return GPR values instead of
+    contract traces, which is used to check for mismatches between the model and the executor """
+
+    def __init__(self,
+                 sandbox_base: int,
+                 code_start: int,
+                 tracer: Tracer,
+                 enable_mismatch_check_mode: bool = False):
+        self.code_start = code_start
+        self.sandbox_base = sandbox_base
+        self.tracer = tracer
+        self.mismatch_check_mode = enable_mismatch_check_mode
+
+        self.instruction_coverage = defaultdict(int)
         super().__init__()
 
     @abstractmethod
@@ -858,64 +1146,125 @@ class Model(ABC):
         pass
 
     @abstractmethod
-    def get_taints(self, inputs, nesting) -> List[InputTaint]:
+    def trace_test_case_with_taints(self, inputs, nesting) -> Tuple[List[CTrace], List[InputTaint]]:
         pass
 
-    def set_coverage(self, coverage: Coverage):
-        self.coverage = coverage
-
 
 class Executor(ABC):
-    coverage: Optional[Coverage] = None
+    mismatch_check_mode: bool = False
+    """ mismatch_check_mode: If True, the executor will return GPR values instead of
+    hardware traces, which is used to check for mismatches between the model and the executor """
+
+    @abstractmethod
+    def __init__(self, enable_mismatch_check_mode: bool = False):
+        self.mismatch_check_mode = enable_mismatch_check_mode
+        super().__init__()
 
     @abstractmethod
     def load_test_case(self, test_case: TestCase):
         pass
 
     @abstractmethod
-    def trace_test_case(self, inputs: List[Input], repetitions: int = 0) -> List[CombinedHTrace]:
+    def trace_test_case(self, inputs: List[Input], n_reps: int) -> List[HTrace]:
+        """ Call the executor kernel module to collect the hardware traces for
+         the test case (previously loaded with `load_test_case`) and the given inputs.
+
+        :param inputs: list of inputs to be used for the test case
+        :param n_reps: number of times to repeat each measurement
+        :return: a list of HTrace objects, one for each input
+         """
         pass
 
     @abstractmethod
     def read_base_addresses(self) -> Tuple[int, int]:
         pass
 
-    def set_coverage(self, coverage: Coverage):
-        self.coverage = coverage
+    @abstractmethod
+    def set_ignore_list(self, ignore_list: List[int]):
+        """ Sets a list of inputs IDs that should be ignored by the executor.
+        The executor will executed the inputs with these IDs as normal (in case they are
+        necessary for priming the uarch state), but their htraces will be set to zero """
+        pass
 
     @abstractmethod
-    def get_last_feedback(self) -> List:
+    def extend_ignore_list(self, ignore_list: List[int]):
+        """ Updates the ignore list with a new list of inputs IDs that should be ignored
+        by the executor."""
         pass
 
 
 class Analyser(ABC):
-    coverage: Optional[Coverage] = None
 
     @abstractmethod
     def filter_violations(self,
                           inputs: List[Input],
                           ctraces: List[CTrace],
                           htraces: List[HTrace],
-                          stats=False) -> List[EquivalenceClass]:
+                          stats=False) -> List[Violation]:
         pass
 
-    def set_coverage(self, coverage: Coverage):
-        self.coverage = coverage
+    @abstractmethod
+    def htraces_are_equivalent(self, htrace1: HTrace, htrace2: HTrace) -> bool:
+        """ Compare two hardware traces according to the current analyser's rules.
+
+        :param htrace1: first hardware trace
+        :param htrace2: second hardware trace
+        :return: True if the traces are equivalent, False otherwise
+        """
+        pass
+
+
+class Fuzzer(ABC):
+    model: Model
+    executor: Executor
+    asm_parser: AsmParser
+    generator: Generator
+    input_gen: InputGenerator
+    analyser: Analyser
+
+    @abstractmethod
+    def initialize_modules(self):
+        pass
+
+    @abstractmethod
+    def filter(self, test_case, inputs) -> bool:
+        """
+        A filter function that can be used to check if a test case is not useful
+        :param test_case: The test case to be checked
+        :param inputs: The inputs to be used with the test case
+        :return: True if the test case should be filtered out (not useful), False otherwise (useful)
+        """
+        pass
+
+    @abstractmethod
+    def fuzzing_round(self,
+                      test_case: TestCase,
+                      inputs: List[Input],
+                      ignore_list: List[int] = []) -> Optional[Violation]:
+        pass
 
 
 class Minimizer(ABC):
 
-    def __init__(self, instruction_set_spec: InstructionSetAbstract):
+    def __init__(self, fuzzer: Fuzzer, instruction_set_spec: InstructionSetAbstract):
         pass
 
     @abstractmethod
-    def minimize(self, test_case_asm: str, outfile: str, num_inputs: int, add_fences: bool):
+    def run(self, test_case_asm: str, n_inputs: int, test_case_outfile: str, input_outdir: str,
+            n_attempts: int, **kwargs):
+        """
+        Run the minimizer on a test case.
+        See postprocessor.py:MainMinimizer for the full list of arguments.
+        """
         pass
 
 
 class TaintTrackerInterface(ABC):
 
-    def __init__(self, initial_observations, sandbox_base=0):
+    def __init__(self, initial_observations: List[str], sandbox_base: int = 0):
+        pass
+
+    def reset(self, initial_observations: List[str]):
         pass
 
     def start_instruction(self, instruction: Instruction) -> None:
@@ -930,10 +1279,7 @@ class TaintTrackerInterface(ABC):
     def taint_memory_access_address(self):
         pass
 
-    def taint_memory_load(self):
-        pass
-
-    def taint_memory_store(self):
+    def taint_loaded_value(self):
         pass
 
     def checkpoint(self):
@@ -945,3 +1291,26 @@ class TaintTrackerInterface(ABC):
     @abstractmethod
     def get_taint(self) -> InputTaint:
         pass
+
+
+# ==================================================================================================
+# Exceptions
+# ==================================================================================================
+class GeneratorException(Exception):
+    """ Exception raised when an error occurs during test case generation """
+    pass
+
+
+class HardwareTracingError(Exception):
+    """ Exception raised when an error occurs during hardware tracing """
+    pass
+
+
+class NotSupportedException(Exception):
+    """ Exception raised when a feature is not supported """
+    pass
+
+
+class UnreachableCode(Exception):
+    """ Exception raised when an unreachable code path is reached """
+    pass
